var gltf = {};

/*
	Requests an url as a json file, then does some processing on that, and finally calls a js or c2 function.
	Expects an "embeded" gltf file.
*/
function loadGltf(url, callback)
{
	var request = new XMLHttpRequest();
	request.onreadystatechange = function()
	{
		if(this.readyState==4)
		{
			if(this.status==200)
			{
				gltf = this.response;
				gltf_cleanup();
				if(callback)
				{
					if(typeof callback =='string') 
						c2_callFunction(callback,[]);
					else 
						callback();
				}
			}
			else
				console.log("error: can't load '"+url+"'.")
		}
	};
	request.responseType = 'json';
	request.open("GET", url, true);
	request.send();
}

/*
	Once the gltf file is loaded as a json this does a few things to it to make out lives easier.
	1. It converts the buffers from base64 to a typed array.
	2. The accessors are then made as typed views of the buffers.
	3. All the places where indexes are used are changed to look at those objects themselves.
	4. Adds some properties that are optionally in the gltf file.
	5. Adds a few properties that are utilized when getting data from skinning.
*/
function gltf_cleanup()
{
	//extra variable for a list of skinned meshes.  They need to be transformed after the rest.
	gltf.skinnedNodes = [];
	
	// buffers
	for(let i in gltf.buffers)  // convert to typed arrays.
	{
		let base64 = gltf.buffers[i].uri.slice(37)
		gltf.buffers[i] = Uint8Array.from(atob(base64), c=>c.charCodeAt(0)).buffer;
	}
	
	// accessors
	for(let i in gltf.accessors)
	{
		let a = gltf.accessors[i];
		let buftype = null;
		switch(a.componentType)
		{
			case 5120: buftype=Int8Array; break;
			case 5121: buftype=Uint8Array; break;
			case 5122: buftype=Int16Array; break;
			case 5123: buftype=Uint16Array; break;
			case 5125: buftype=Uint32Array; break;
			case 5126: buftype=Float32Array; break;
			default: console.log("error: gltf, unhandled componentType");
		}
		let compcount = {"SCALAR":1, "VEC2":2, "VEC3":3, "VEC4":4, "MAT2":4, "MAT3":9, "MAT4":16}[a.type];
		let bufview = gltf.bufferViews[a.bufferView];
		a.data = new buftype(gltf.buffers[bufview.buffer], bufview.byteOffset, compcount*a.count);
	}
	
	// scene
	gltf.scene = gltf.scenes[gltf.scene];
	
	// scenes
	for(let i in gltf.scenes)
	{
		let s = gltf.scenes[i];
		for(let j in s.nodes)
			s.nodes[j] = gltf.nodes[s.nodes[j]];
	}
	
	// nodes
	for(let i in gltf.nodes)
	{
		let n = gltf.nodes[i];
		
		if(n.translation == undefined)
			n.translation = [0,0,0];
		if(n.scale == undefined)
			n.scale = [1,1,1];
		if(n.rotation == undefined)
			n.rotation = [0,0,0,1];
		if(n.matrix == undefined)
			n.matrix = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
		
		if(n.mesh != undefined)
			n.mesh = gltf.meshes[n.mesh];
		if(n.skin != undefined)
		{
			n.skin = gltf.skins[n.skin];
			n.invMatrix = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
			gltf.skinnedNodes.push(n);
		}
		if(n.children != undefined)
			for(let j in n.children)
				n.children[j] = gltf.nodes[n.children[j]];
	}
	
	// animations
	for(let i in gltf.animations)
	{
		let a = gltf.animations[i];
		
		for(let j in a.channels)
		{
			let c = a.channels[j];
			c.sampler = a.samplers[c.sampler];
			c.target.node = gltf.nodes[c.target.node];
		}
		
		for(let j in a.samplers)
		{
			let s = a.samplers[j];
			s.input = gltf.accessors[s.input];
			s.output = gltf.accessors[s.output];
		}
	}
	
	//meshes
	for(let i in gltf.meshes)
	{
		let m = gltf.meshes[i];
		for(let j in m.primitives)
		{
			let p = m.primitives[j];
			
			p.indices = gltf.accessors[p.indices];
			p.material = gltf.materials[p.material];
			
			Object.keys(p.attributes).forEach(function(key){
				p.attributes[key] = gltf.accessors[p.attributes[key]];
			});
		}
	}
	
	//skins
	for(let i in gltf.skins)
	{
		let s = gltf.skins[i];
		s.inverseBindMatrices = gltf.accessors[s.inverseBindMatrices];
		for(let j in s.joints)
		{
			s.joints[j] = gltf.nodes[s.joints[j]];
			s.joints[j].invBindMatrix = s.inverseBindMatrices.data.subarray(j*16, j*16+16);
			s.joints[j].boneMatrix = [0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0];
		}
	}
}

// draw points with c2's renderer directly
function hijackC2Sprite(uid)
{
	gltf.pointBatch = [];
	let obj = cr_getC2Runtime().getObjectByUID(uid);
	obj.oldDrawGL = obj.drawGL;
	obj.drawGL = function(glw)
	{
		this.oldDrawGL(glw);
		
		let batch = gltf.pointBatch;
		let len = batch.length/3
		let x,y,z;
		for(let i=0; i<len; i++)
		{
			x=batch[i*3+0];
			y=batch[i*3+1];
			z=0.25-batch[i*3+2];
			glw.point(320+320*x/z, 350-320*y/z, 32*0.03/z, 0.1/z);
		}
		batch.length=0;
	};
}

/*
	Updates a node's matrix and all it's children nodes.
	After that it transforms unskinned mesh points and sends them to c2.
*/
function transformNode(node, parentMat)
{
	if(parentMat != undefined)
		mat4.set(parentMat, node.matrix);
	else
		mat4.identity(node.matrix);
	mat4.translate(node.matrix, node.translation);
	mat4.multiply(node.matrix, quat4.toMat4(node.rotation));
	mat4.scale(node.matrix, node.scale);
	
	if(node.skin != undefined)
		mat4.inverse(node.matrix, node.invMatrix);
	
	// unskinned meshes
	if(node.mesh != undefined && node.skin == undefined)  
	{
		let transformedVerts = [];
		
		for(let i in node.mesh.primitives)
		{
			transformedVerts.length = 0;
			let posData = node.mesh.primitives[i].attributes.POSITION.data;
			
			let v = [0,0,0];			
			for(let j=0; j<posData.length/3; j++)
			{
				mat4.multiplyVec3(node.matrix, posData.subarray(j*3, j*3+3), v);
				transformedVerts.push(v[0],v[1],v[2]);
			}
			
			if(transformedVerts.length > 0)
			{
				if(gltf.pointBatch != undefined)
					for(let ii in transformedVerts)
						gltf.pointBatch.push(transformedVerts[ii]);
				else
				{
					c2_callFunction('points', transformedVerts);
					c2_callFunction('texcoords', node.mesh.primitives[i].attributes.TEXCOORD_0.data);
					c2_callFunction('indices', node.mesh.primitives[i].indices.data);
				}
			}
		}
	}
	
	if(node.children != undefined)
		for(let i in node.children)
			transformNode(node.children[i], node.matrix);
}

//	Updates scene graph, and as a second step sends transformed skinned mesh points to c2.
function getPoints()
{
	// update all scene matrixes.
	for(let i in gltf.scene.nodes)
		transformNode(gltf.scene.nodes[i]);
	
	//todo loop over skinned nodes.
	//todo: limit to ones in scene?
	for(let ii in gltf.skinnedNodes)
	{
		let node = gltf.skinnedNodes[ii];
		
		//update bone matrixes
		for(let jj in node.skin.joints)
		{
			let joint = node.skin.joints[jj];
			
			mat4.multiply(node.invMatrix, joint.matrix, joint.boneMatrix);
			mat4.multiply(joint.boneMatrix, joint.invBindMatrix);
		}
		
		let transformedVerts = [];
		
		for(let i in node.mesh.primitives)
		{
			transformedVerts.length=0;
			
			let posData = node.mesh.primitives[i].attributes.POSITION.data;
			let weights = node.mesh.primitives[i].attributes.WEIGHTS_0.data;
			let joints = node.mesh.primitives[i].attributes.JOINTS_0.data;
			
			for(let j=0; j<posData.length/3; j++)
			{
				let w = weights.subarray(j*4, j*4+4);
				let b = joints.subarray(j*4, j*4+4);
				let vin = posData.subarray(j*3, j*3+3)
				let v = [0,0,0], vsum = [0,0,0];
				
				for(let i=0; i<4; i++)
				{
					mat4.multiplyVec3(node.skin.joints[b[i]].boneMatrix, vin, v);
					vec3.scale(v, w[i]);
					vec3.add(vsum, v);
				}
				
				transformedVerts.push(vsum[0],vsum[1],vsum[2]);
			}
			
			if(transformedVerts.length > 0)
			{
				if(gltf.pointBatch != undefined)
					for(let ii in transformedVerts)
						gltf.pointBatch.push(transformedVerts[ii]);
				else
				{
					c2_callFunction('points', transformedVerts);
					c2_callFunction('texcoords', node.mesh.primitives[i].attributes.TEXCOORD_0.data);
					c2_callFunction('indices', node.mesh.primitives[i].indices.data);
				}
			}
		}
	}
}

// sends a list of animation names to c2.
function getAnimationNames()
{
	let names = [];
	
	for(let i in gltf.animations)
		names.push(gltf.animations[i].name);
	
	c2_callFunction("animations", names);
}

// quat.slerp is broken in the version of glmatrix included in c2, so replace it with a newer one. 
quat4.slerp = function(a, b, t, out) 
{
	var ax = a[0],
	ay = a[1],
	az = a[2],
	aw = a[3];
	var bx = b[0],
	by = b[1],
	bz = b[2],
	bw = b[3];
	var omega, cosom, sinom, scale0, scale1;

	cosom = ax * bx + ay * by + az * bz + aw * bw;

	if (cosom < 0.0) {
		cosom = -cosom;
		bx = -bx;
		by = -by;
		bz = -bz;
		bw = -bw;
	}

	if (1.0 - cosom > 0.000001) {
		omega = Math.acos(cosom);
		sinom = Math.sin(omega);
		scale0 = Math.sin((1.0 - t) * omega) / sinom;
		scale1 = Math.sin(t * omega) / sinom;
	} else {
		scale0 = 1.0 - t;
		scale1 = t;
	}

	out[0] = scale0 * ax + scale1 * bx;
	out[1] = scale0 * ay + scale1 * by;
	out[2] = scale0 * az + scale1 * bz;
	out[3] = scale0 * aw + scale1 * bw;
	return out;
}

// Updates animation at index to be at time.  Is used to play animation.  
function updateAnimation(index, time)
{
	let anim = gltf.animations[index];
	
	for(let i in anim.channels)
	{
		let c = anim.channels[i];
		let timeValues = c.sampler.input;
		let otherValues = c.sampler.output.data;
		let target = c.target;
		
		time = (time-timeValues.min[0])%(timeValues.max[0]-timeValues.min[0]) + timeValues.min[0]; // loop
		//time = Math.min(Math.max(time, timeValues.min[0]), timeValues.max[0]);  //clamp
		timeValues = timeValues.data;
		
		//find index in timeline
		let t0, t1;
		for(t0=0, t1=1; t0<timeValues.length-1; t0++, t1++)  
			if(time >= timeValues[t0] && time <= timeValues[t1])
				break;
		let t = (time - timeValues[t0])/(timeValues[t1] - timeValues[t0]);
		
		if(target.path == "translation")
			vec3.lerp(otherValues.subarray(t0*3,t0*3+3), otherValues.subarray(t1*3,t1*3+3), t, target.node.translation);
		else if(target.path == "rotation")
			quat4.slerp(otherValues.subarray(t0*4,t0*4+4), otherValues.subarray(t1*4,t1*4+4), t, target.node.rotation);
		else if(target.path == "scale")
			vec3.lerp(otherValues.subarray(t0*3,t0*3+3), otherValues.subarray(t1*3,t1*3+3), t, target.node.scale);
	}
}